import signal
from pathlib import Path
from threading import Thread
from time import sleep
from typing import Optional

import rich
from click import clear
from rich.align import Align

from abst.bastion_support.oci_bastion import Bastion
from abst.config import default_stack_location, default_stack_contents, \
    default_contexts_location
from abst.wrappers import load_stack_decorator


class BastionScheduler:
    __dry_stack = set()
    __live_stack = set()
    stopped = False
    session_list = []

    @classmethod
    def load_stack(cls):
        if not default_stack_location.exists():
            Bastion.write_creds_json(default_stack_contents, default_stack_location)
        if default_stack_location.exists():
            cls.__dry_stack = set(Bastion.load_json(default_stack_location)["stack"])

    @classmethod
    def update_stack(cls):
        if default_stack_location.exists():
            Bastion.write_creds_json({"stack": list(cls.__dry_stack)},
                                     default_stack_location)

    @classmethod
    def get_contexts(cls):
        return [name.name.replace(".json", "") for name in
                default_contexts_location.iterdir()]

    @classmethod
    @load_stack_decorator
    def add_bastion(cls, context_name: str):
        if context_name in cls.__dry_stack:
            print("Already in stack")
            return

        if context_name != "default" and context_name not in cls.get_contexts():
            print(f"No context with name {context_name}")
            return

        if (res := cls.local_port_in_stack(context_name))[0]:
            rich.print(
                f"[red]Local port is already taken by {res[1].context_name}[/red]")
            return

        cls.__dry_stack.add(context_name)
        rich.print(f"[green]Added Bastion {context_name}[/green]")
        cls.update_stack()

    @classmethod
    @load_stack_decorator
    def remove_bastion(cls, name: str):
        try:
            cls.__dry_stack.remove(name)
            rich.print(f"[green]Removed Bastion {name}[/green]")
            cls.update_stack()
        except KeyError:
            rich.print(f"[red]No Bastion by name {name} in stack[/red]")

    @classmethod
    def __display(cls, console):
        console.clear()
        from rich.table import Table
        table = Table(title="Bastion Sessions", highlight=True)
        table.add_column("Name", justify="left", style="cyan", no_wrap=True)
        table.add_column("Local Port", style="magenta", no_wrap=True)
        table.add_column("Active", justify="right", style="green", no_wrap=True)
        table.add_column("Status", justify="right", style="green", no_wrap=True)
        for bastion in cls.__live_stack:
            conf = bastion.load_self_creds()
            active = bastion.connected and bastion.active_tunnel.poll() is None
            name = "default" if bastion.context_name is None else bastion.context_name

            try:
                table.add_row(name, conf.get('local-port', 'Not Specified'), str(active),
                              bastion.current_status)
            except:
                pass
        centered_table = Align.center(table)
        console.print(centered_table)

    @classmethod
    def __display_loop(cls):
        from rich.console import Console
        console = Console()
        signal.signal(signal.SIGINT, BastionScheduler.kill_all)
        signal.signal(signal.SIGTERM, BastionScheduler.kill_all)
        while True:
            clear()
            cls.__display(console)
            sleep(3)

    @classmethod
    def _run_indefinitely(cls, func, force: bool = False):
        while True:
            if cls.stopped:
                return
            try:
                func(force=force)
            finally:
                sleep(1)

    @classmethod
    @load_stack_decorator
    def run(cls, force=False, set_dir: Optional[Path] = None):
        signal.signal(signal.SIGINT, BastionScheduler.kill_all)
        signal.signal(signal.SIGTERM, BastionScheduler.kill_all)
        rich.print("Will run all Bastions in parallel")
        thread_list = []

        if not set_dir:
            for context_name in cls.__dry_stack:
                if cls.stopped:
                    return
                bastion = Bastion(None if context_name == "default" else context_name, region=
                Bastion.load_json(Bastion.get_creds_path_resolve(context_name)).get("region", None))
                Bastion.session_list.append(bastion)
                cls.session_list.append(bastion)
                cls.__live_stack.add(bastion)
                t = Thread(name=context_name, target=cls._run_indefinitely,
                           args=[bastion.create_forward_loop, force], daemon=True)
                thread_list.append(t)
                t.start()
                rich.print(f"Started {context_name}")
        else:
            for context_path in filter(lambda p: not str(p.name).startswith("."), set_dir.iterdir()):
                if cls.stopped:
                    return
                context_name = context_path.name[:-5]
                bastion = Bastion(None if context_name == "default" else context_name, region=
                Bastion.load_json(context_path).get("region", None), direct_json_path=context_path)
                Bastion.session_list.append(bastion)
                cls.session_list.append(bastion)
                cls.__live_stack.add(bastion)
                t = Thread(name=context_name, target=cls._run_indefinitely,
                           args=[bastion.create_forward_loop, force], daemon=True)
                thread_list.append(t)
                t.start()
                rich.print(f"Started {context_name}")


        cls.__display_loop()

    @classmethod
    @load_stack_decorator
    def local_port_in_stack(cls, context_name: str):
        path = Bastion.get_creds_path_resolve(context_name)
        instance_creds = Bastion.load_json(path)
        for context_name in cls.__dry_stack:
            b_creds = Bastion.load_json(Bastion.get_creds_path_resolve(context_name))
            if b_creds["local-port"] == instance_creds["local-port"]:
                return True, context_name
        return False, None

    @classmethod
    def kill_all(cls, a=None, b=None, c=None):
        # This should be only executed in running state
        cls.stopped = True
        blist_copy = list(cls.session_list)
        for i, sess in enumerate(blist_copy):
            try:
                print(f"Killing {sess.bid}")
                Bastion.current_status = "deleting"
                sess.kill()
                Bastion.delete_bastion_session(sess.bid, sess.region)
            except Exception:
                print(f"Looks like Bastion is already deleted {sess}")
            finally:
                print(f"Deleting {i + 1}/{len(blist_copy)}")
        exit(0)

    @classmethod
    @load_stack_decorator
    def get_bastions(cls) -> tuple:
        return tuple(cls.__dry_stack)



import click
import rich
from InquirerPy import inquirer

from abst.bastion_support.bastion_scheduler import BastionScheduler
from abst.config import default_parallel_sets_location
from abst.tools import display_scheduled
from abst.utils.misc_funcs import setup_calls


@click.group(help="Parallel Bastion Control group")
def parallel():
    """
    Only Port Forwarded sessions are supported

    This makes it possible to run multiple forward sessions of multiple context in
     the same time, useful when working with multiple clusters
    """
    pass


@parallel.command("add", help="Add Bastion to stack")
@click.option("--debug", is_flag=True, default=False)
@click.argument("context-name", default="default")
def add(debug, context_name):
    """
    Add Bastions to stack

    use default for adding default cluster ! this is reserved name!
    :param debug:
    :param context_name:
    :return:
    """
    setup_calls(debug)

    BastionScheduler.add_bastion(context_name)
    display_scheduled()


@parallel.command("create", help="Create folder for a parallel set")
@click.option("--debug", is_flag=True, default=False)
@click.argument("set-name", default="default")
def create(debug, set_name):
    """
    Create folder for a parallel set

    :param debug:
    :param set_name:
    :return:
    """
    setup_calls(debug)
    try:
        set_dir = default_parallel_sets_location / set_name
        set_dir.mkdir()
        rich.print(f"[green]Set {set_name} created![/green]")
        rich.print(
            f"Feel free to copy any context to {set_dir} for it to be run in [yellow]{set_name}[/yellow] parallel set")
        rich.print(f"You can run this set by 'abst parallel run {set_name}'")
    except OSError:
        rich.print("[red]Set already exists[/red]")


@parallel.command("list", help="List folder of a parallel set")
@click.option("--debug", is_flag=True, default=False)
def _list(debug):
    setup_calls(debug)
    rich.print("Sets in parallel folder")
    for _set in default_parallel_sets_location.iterdir():
        if _set.name.startswith("."):
            continue
        rich.print(f"   {_set.name}")
        for ctx in _set.iterdir():
            if ctx.name.startswith("."):
                continue
            rich.print(f"       {ctx.name.replace('.json', '')}")


@parallel.command("remove", help="Remove Bastion from stack (Can not remove sets)")
@click.option("--debug", is_flag=True, default=False)
@click.argument("context-name", default="default")
def remove(debug, context_name):
    setup_calls(debug)
    BastionScheduler.remove_bastion(context_name)
    display_scheduled()


@parallel.command("run", help="Run All Bastions in fullauto")
@click.option("--debug", is_flag=True, default=False)
@click.option("-y", is_flag=True, default=False, help="Automatically confirm")
@click.option("-f", "--force", is_flag=True, default=False, help="Will force connections ignoring security policies")
@click.argument("set_name", default=None, required=False, type=str)
def run(debug, y, force, set_name=None):
    setup_calls(debug)
    if force:
        rich.print("[red]Running in force mode[/red][gray] this mode is less secure as it is ignoring key checking and"
                   " security policies involving known_hosts[/gray]")
    if set_name:
        set_dir = get_set_dir(set_name)
    else:
        set_dir = None

    display_scheduled(set_dir)
    if not y:
        try:
            confirm = inquirer.confirm(
                "Do you really want to run following contexts?"
            ).execute()
        except KeyError:
            rich.print("Unknown inquirer error, continuing...")
            confirm = True
        if not confirm:
            rich.print("[green]Cancelling, nothing started[/green]")
            exit(0)
    BastionScheduler.run(force, set_dir)


def get_set_dir(set_name):
    set_dir = default_parallel_sets_location / set_name
    if not set_dir.exists():
        rich.print(f"Parallel set {set_name} did not found in {set_dir}")
        exit(1)
    elif len(list(set_dir.iterdir())) == 0:
        rich.print(f"[red]No contexts found in {set_dir}[/red]")
        exit(1)
    return set_dir


@parallel.command("display", help="Display current Bastions is stack")
@click.option("--debug", is_flag=True, default=False)
@click.argument("set_name", default=None, required=False, type=str)
def display(debug, set_name):
    setup_calls(debug)
    set_dir = get_set_dir(set_name)
    display_scheduled(set_dir)



import click
import rich
from InquirerPy import inquirer

from abst.bastion_support.oci_bastion import Bastion
from abst.cfg_func import __upgrade
from abst.config import default_contexts_location
from abst.tools import get_context_path
from abst.utils.misc_funcs import setup_calls


@click.group(help="Group of commands for operations with config")
def config():
    pass


@config.command("generate", help="Will generate sample json and overwrite changes")
@click.option("--debug", is_flag=True, default=False)
@click.argument("context-name", default=None, required=False)
def generate(debug, context_name):
    setup_calls(debug)

    path = get_context_path(context_name)

    Bastion.create_default_locations()
    td = Bastion.generate_sample_dict()
    creds_path = Bastion.write_creds_json(td, path)
    print(
        f"Sample credentials generated, please fill 'creds.json' in {creds_path} with "
        f"your credentials for this to work, you can use 'abst json fill "
        f"{context_name if context_name else ''}'"
    )


@config.command(
    "fill", help="Fills Json config with credentials you enter interactively"
)
@click.option("--debug", is_flag=True, default=False)
@click.argument("context-name", default=None, required=False)
def fill(debug, context_name):
    setup_calls(debug)

    path = get_context_path(context_name)

    if not path.exists():
        rich.print("Generating sample Creds file")
        Bastion.create_default_locations()
        td = Bastion.generate_sample_dict()
        Bastion.write_creds_json(td, path)

    if not default_contexts_location.exists():
        rich.print("Generating contexts location")
        Bastion.create_default_locations()

    rich.print(f"[green]Filling {str(path)}")
    rich.print("Please fill field one by one as displayed")
    n_dict = dict()

    creds_json_ro = Bastion.load_json(path)

    for key, value in creds_json_ro.items():
        n_dict[key] = inquirer.text(
            message=f"{key.capitalize()}:", default=value
        ).execute()
    rich.print("\n[red]New json looks like this:[/red]")
    rich.print_json(data=n_dict)
    if inquirer.confirm(message="Write New Json ?", default=False).execute():
        Bastion.write_creds_json(n_dict, path)
        rich.print("[green]Wrote changes[/green]")
    else:
        rich.print("[red]Fill interrupted, nothing changed[/red]")


@config.command(
    "locate", help="Locates Json config with credentials you enter interactively"
)
@click.option("--debug", is_flag=True, default=False)
@click.argument("context-name", default=None, required=False)
def locate(debug, context_name):
    setup_calls(debug)

    path = get_context_path(context_name)

    if path.exists():
        rich.print(f"[green]Config file location: {path.absolute()}[/green]")
    else:
        rich.print(
            f"[red]Config does not exist yet, future location"
            f" {path.absolute()}[/red]"
        )


@config.command(
    "upgrade", help="Upgrades Json config with credentials you enter interactively"
)
@click.option("--debug", is_flag=True, default=False)
@click.option("--all", is_flag=True, default=False)
@click.argument("context-name", default=None, required=False)
def upgrade(debug, context_name, all):
    setup_calls(debug)

    path = get_context_path(context_name)

    if not path.exists():
        rich.print("[green]No config to upgrade[/green]")
        return

    __upgrade(context_name, path, all)



import rich
from InquirerPy import inquirer

from abst.config import default_conf_path


def helm_login_dialog():
    rich.print(f"Please Fill in Repository details that"
               f" are going to be saved in {default_conf_path}")
    host = inquirer.text("Host", default="phx.ocir.io").execute()
    remote = inquirer.text("Remote URL", default="oci://").execute()
    username = inquirer.text("Username").execute()
    password = inquirer.secret("Password").execute()
    return host, password, remote, username



import os
import subprocess

import click
import rich
from InquirerPy import inquirer

from abst.bastion_support.oci_bastion import Bastion
from abst.config import default_conf_path
from abst.dialogs import helm_login_dialog
from abst.utils.misc_funcs import setup_calls


@click.group("helm", help="Commands for helm and docker")
def helm():
    pass


@helm.command("login")
@click.option("--debug", is_flag=True, default=False)
@click.option(
    "--edit",
    default=False,
    help="Edit credentials with index as argument (starts from 1)",
)
def helm_login(debug, edit):
    """
    Copy Secret in current cluster from source namespace to target
    @return:
    :param debug:
    """
    setup_calls(debug)
    from subprocess import PIPE

    try:
        _config = Bastion.load_config()
        if "helm" not in _config.keys():
            _config["helm"] = []
            host, password, remote, username = helm_login_dialog()
            _config["helm"].append(
                {
                    "host": host,
                    "remote": remote,
                    "username": username,
                    "password": password,
                }
            )
            Bastion.write_creds_json(_config, default_conf_path)
            rich.print("Added Credentials")
        elif edit:
            if "helm" not in _config.keys():
                _config["helm"] = []

            host, password, remote, username = helm_login_dialog()
            _config["helm"][edit - 1] = {
                "host": host,
                "remote": remote,
                "username": username,
                "password": password,
            }
            rich.print_json(data=_config["helm"][edit - 1])
            if inquirer.confirm(message="Write New Json ?", default=False).execute():
                Bastion.write_creds_json(_config, default_conf_path)
                rich.print("Edited Credentials")

        choices = [
            {"name": f'{choice["remote"]} {choice["username"]}', "value": choice}
            for choice in _config["helm"]
        ]

        selected = inquirer.select("Select login", choices).execute()
        rich.print("Trying to login")
        p = subprocess.Popen(
            f'helm registry login {selected["host"]} -u {selected["username"]} --password-stdin'.split(
                " "
            ),
            stdout=PIPE,
            stdin=PIPE,
            stderr=PIPE,
        )
        rich.print(p.communicate(input=selected["password"].encode())[0].decode())

    except Exception as ex:
        rich.print(f"[red]Exception during execution {ex}[/red]")
        return


@helm.command("push")
@click.argument("chart")
@click.option("--debug", is_flag=True, default=False)
def helm_push(chart, debug):
    """
    Copy Secret in current cluster from source namespace to target
    @return:
    :param debug:
    :param chart:
    :param debug:
    """
    setup_calls(debug)
    try:
        _config = Bastion.load_config()
        if "helm" not in _config.keys():
            _config["helm"] = []
            rich.print(
                f"Please Fill in Repository details that"
                f" are going to be saved in {default_conf_path}"
                f" with command 'abst helm login'"
            )
        choices = [
            {"name": f'{choice["remote"]} {choice["username"]}', "value": choice}
            for choice in _config["helm"]
        ]

        selected = inquirer.select("Select remote", choices).execute()
        rich.print("Trying to push")
        os.system(f'helm push {chart} {selected["remote"]}')

    except Exception as ex:
        rich.print(f"[red]Exception during execution {ex}[/red]")
        return
